import re
import tempfile
from pathlib import Path
from typing import Iterable

from PIL.Image import Image

from ..common import ASSET_TAGS, AssetHub, remove_surrogates
from ..error import OCRError
from ..metering import AbortedCheck, check_aborted
from .ngrams import has_repetitive_ngrams
from .types import DeepSeekOCRSize, Page, PageLayout


class PageExtractorNode:
    def __init__(
        self,
        model_path: Path | None = None,
        local_only: bool = False,
        enable_devices_numbers: Iterable[int] | None = None,
    ) -> None:
        self._model_path: Path | None = model_path
        self._local_only: bool = local_only
        self._enable_devices_numbers: Iterable[int] | None = enable_devices_numbers
        self._page_extractor = None

    def _get_page_extractor(self):
        if not self._page_extractor:
            # 尽可能推迟 doc-page-extractor 的加载时间
            from doc_page_extractor import create_page_extractor

            self._page_extractor = create_page_extractor(
                model_path=self._model_path,
                local_only=self._local_only,
                enable_devices_numbers=self._enable_devices_numbers,
            )
        return self._page_extractor

    def download_models(self, revision: str | None) -> None:
        self._get_page_extractor().download_models(revision)

    def load_models(self) -> None:
        self._get_page_extractor().load_models()

    def image2page(
        self,
        image: Image,
        page_index: int,
        asset_hub: AssetHub,
        ocr_size: DeepSeekOCRSize,
        includes_footnotes: bool,
        includes_raw_image: bool,
        plot_path: Path | None,
        max_tokens: int | None,
        max_output_tokens: int | None,
        device_number: int | None,
        aborted: AbortedCheck,
    ) -> Page:
        from doc_page_extractor import ExtractionContext, plot

        body_layouts: list[PageLayout] = []
        footnotes_layouts: list[PageLayout] = []
        raw_image: Image | None = None

        if includes_raw_image:
            raw_image = image
            image = image.copy()

        with tempfile.TemporaryDirectory() as temp_dir_path:
            context = ExtractionContext(
                check_aborted=aborted,
                max_tokens=max_tokens,
                max_output_tokens=max_output_tokens,
                output_dir_path=temp_dir_path,
            )
            step_index: int = 1
            generator = self._get_page_extractor().extract(
                image=image,
                size=ocr_size,
                stages=2 if includes_footnotes else 1,
                context=context,
                device_number=device_number,
            )
            while True:
                try:
                    image, layouts = next(generator)
                except StopIteration:
                    break
                except Exception as error:
                    raise OCRError(
                        f"Failed to extract page {page_index} layout at stage {step_index}.",
                        page_index=page_index,
                        step_index=step_index,
                    ) from error

                for layout in layouts:
                    ref = self._normalize_text(layout.ref)
                    text = self._normalize_text(layout.text)
                    det = self._normalize_layout_det(image.size, layout.det)

                    if det is None:
                        continue

                    # 检测短模式重复（如 "1.1.1.1."）
                    if has_repetitive_ngrams(
                        text, min_ngram=2, max_ngram=5, repeat_threshold=16
                    ):
                        continue

                    # 检测长模式重复（保守策略）
                    if has_repetitive_ngrams(
                        text, min_ngram=6, max_ngram=20, repeat_threshold=8
                    ):
                        continue

                    hash: str | None = None
                    if ref in ASSET_TAGS:
                        hash = asset_hub.clip(image, det)

                    if step_index == 1:
                        order = len(body_layouts)
                    elif step_index == 2 and ref not in ASSET_TAGS:
                        order = len(footnotes_layouts)
                    else:
                        continue

                    page_layout = PageLayout(
                        ref=ref,
                        det=det,
                        text=text,
                        hash=hash,
                        order=order,
                    )
                    if step_index == 1:
                        body_layouts.append(page_layout)
                    elif step_index == 2 and ref not in ASSET_TAGS:
                        footnotes_layouts.append(page_layout)

                check_aborted(aborted)
                if plot_path is not None:
                    plot_file_path = (
                        plot_path / f"page_{page_index}_stage_{step_index}.png"
                    )
                    image = plot(image.copy(), layouts)
                    image.save(plot_file_path, format="PNG")
                    check_aborted(aborted)

                step_index += 1

            return Page(
                index=page_index,
                image=raw_image,
                body_layouts=body_layouts,
                footnotes_layouts=footnotes_layouts,
                input_tokens=context.input_tokens,
                output_tokens=context.output_tokens,
            )

    def _normalize_text(self, text: str | None) -> str:
        if text is None:
            return ""
        text = remove_surrogates(text)
        text = re.sub(r"\s+", " ", text)
        return text.strip()

    def _normalize_layout_det(
        self,
        size: tuple[int, int],
        det: tuple[int, int, int, int],
    ) -> tuple[int, int, int, int] | None:
        width, height = size
        left, top, right, bottom = det
        left = max(0, min(left, width))
        top = max(0, min(top, height))
        right = max(0, min(right, width))
        bottom = max(0, min(bottom, height))

        if left >= right or top >= bottom:
            return None
        return left, top, right, bottom



import io
import re
from dataclasses import dataclass
from enum import Enum, auto
from typing import Generator, Iterable


class NumberClass(Enum):
    ROMAN_NUMERAL = auto()  # 罗马数字
    LOWERCASE_ROMAN_NUMERAL = auto()  # 小写罗马数字
    CIRCLED_NUMBER = auto()  # 带圆圈的数字
    PARENTHESIZED_CHINESE = auto()  # 括号中的汉字
    CIRCLED_CHINESE = auto()  # 带圆圈的汉字
    BLACK_CIRCLED_NUMBER = auto()  # 黑色圆圈数字
    UNBOUNDED_NUMBER = auto()  # 无包围数字


class NumberStyle(Enum):
    ROMAN_NUMERAL = auto()  # 罗马数字
    LOWERCASE_ROMAN_NUMERAL = auto()  # 小写罗马数字
    CIRCLED_NUMBER = auto()  # 带圆圈的数字
    DOUBLE_CIRCLED_NUMBER = auto()  # 双圈数字
    CIRCLED_SANS_SERIF_NUMBER = auto()  # 带圆圈的无衬线数字
    BLACK_CIRCLED_SANS_SERIF_NUMBER = auto()  # 黑色圆圈无衬线数字
    BLACK_CIRCLED_NUMBER = auto()  # 黑色圆圈数字
    PARENTHESIZED_CHINESE = auto()  # 括号中的汉字
    CIRCLED_CHINESE = auto()  # 带圆圈的汉字
    FULL_WIDTH_NUMBER = auto()  # 全角数字
    MATHEMATICAL_BOLD_NUMBER = auto()  # 数学粗体数字
    ARTISTIC_BOLD_NUMBER = auto()  # 艺术粗体数字
    OUTLINED_BOLD_NUMBER = auto()  # 描边的粗体数字
    SUBSCRIPT_NUMBER = auto()  # 带角标的数字


@dataclass
class Mark:
    number: int
    char: str
    clazz: NumberClass
    style: NumberStyle

    def __str__(self) -> str:
        return self.char

    def __hash__(self):
        return hash((self.clazz, self.number))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Mark):
            return False
        if self.clazz != other.clazz:
            return False
        if self.number != other.number:
            return False
        return True


def samples(number_style: NumberStyle, count: int) -> str:
    if count <= 1:
        raise ValueError("Count must be greater than 1")
    half_count = count // 2
    number_styles = _number_marks.styles.get(number_style, None)
    if number_styles is None:
        raise ValueError(f"Invalid number style: {number_style.name}")
    buffer = io.StringIO()
    for char in number_styles[:half_count]:
        buffer.write(char)
    buffer.write("...")
    for char in number_styles[-half_count:]:
        buffer.write(char)
    return buffer.getvalue()


def transform2mark(raw_char: str) -> Mark | None:
    gotten = _number_marks.marks.get(raw_char, None)
    if gotten is None:
        return None
    return Mark(
        number=gotten.number, char=gotten.char, clazz=gotten.clazz, style=gotten.style
    )


def search_marks(text: str) -> Generator[Mark | str, None, None]:
    for part in re.split(_number_marks.pattern, text):
        mark = transform2mark(part)
        if mark is None:
            yield part
        else:
            yield mark


class _NumberMarks:
    def __init__(
        self,
        styles: Iterable[tuple[NumberClass, NumberStyle, Iterable[tuple[int, str]]]],
    ):
        self.marks: dict[str, Mark] = {}
        self.styles: dict[NumberStyle, list[str]] = {}
        for clazz, style, marks in styles:
            for number, mark in marks:
                self.marks[mark] = Mark(number, mark, clazz, style)
                self.styles[style] = [
                    char for _, char in sorted(marks, key=lambda x: x[0])
                ]

        self.pattern: re.Pattern = re.compile(
            r"([" + "".join(sorted(list(self.marks.keys()))) + r"])"
        )


# some of they are from https://tw.piliapp.com/symbol/number/
_number_marks = _NumberMarks(
    (
        (
            NumberClass.ROMAN_NUMERAL,
            NumberStyle.ROMAN_NUMERAL,
            (
                (1, "Ⅰ"),
                (2, "Ⅱ"),
                (3, "Ⅲ"),
                (4, "Ⅳ"),
                (5, "Ⅴ"),
                (6, "Ⅵ"),
                (7, "Ⅶ"),
                (8, "Ⅷ"),
                (9, "Ⅸ"),
                (10, "Ⅹ"),
                (11, "Ⅺ"),
                (12, "Ⅻ"),
            ),
        ),
        (
            NumberClass.LOWERCASE_ROMAN_NUMERAL,
            NumberStyle.LOWERCASE_ROMAN_NUMERAL,
            (
                (1, "ⅰ"),
                (2, "ⅱ"),
                (3, "ⅲ"),
                (4, "ⅳ"),
                (5, "ⅴ"),
                (6, "ⅵ"),
                (7, "ⅶ"),
                (8, "ⅷ"),
                (9, "ⅸ"),
                (10, "ⅹ"),
                (11, "ⅺ"),
                (12, "ⅻ"),
            ),
        ),
        (
            NumberClass.CIRCLED_NUMBER,
            NumberStyle.CIRCLED_NUMBER,
            (
                (0, "⓪"),
                (1, "①"),
                (2, "②"),
                (3, "③"),
                (4, "④"),
                (5, "⑤"),
                (6, "⑥"),
                (7, "⑦"),
                (8, "⑧"),
                (9, "⑨"),
                (10, "⑩"),
                (11, "⑪"),
                (12, "⑫"),
                (13, "⑬"),
                (14, "⑭"),
                (15, "⑮"),
                (16, "⑯"),
                (17, "⑰"),
                (18, "⑱"),
                (19, "⑲"),
                (20, "⑳"),
                (21, "㉑"),
                (22, "㉒"),
                (23, "㉓"),
                (24, "㉔"),
                (25, "㉕"),
                (26, "㉖"),
                (27, "㉗"),
                (28, "㉘"),
                (29, "㉙"),
                (30, "㉚"),
                (31, "㉛"),
                (32, "㉜"),
                (33, "㉝"),
                (34, "㉞"),
                (35, "㉟"),
                (36, "㊱"),
                (37, "㊲"),
                (38, "㊳"),
                (39, "㊴"),
                (40, "㊵"),
                (41, "㊶"),
                (42, "㊷"),
                (43, "㊸"),
                (44, "㊹"),
                (45, "㊺"),
                (46, "㊻"),
                (47, "㊼"),
                (48, "㊽"),
                (49, "㊾"),
                (50, "㊿"),
            ),
        ),
        (
            NumberClass.CIRCLED_NUMBER,
            NumberStyle.DOUBLE_CIRCLED_NUMBER,
            (
                (0, "⓵"),
                (1, "⓶"),
                (2, "⓷"),
                (3, "⓸"),
                (4, "⓹"),
                (5, "⓺"),
                (6, "⓻"),
                (7, "⓼"),
                (8, "⓽"),
                (9, "⓾"),
            ),
        ),
        (
            NumberClass.CIRCLED_NUMBER,
            NumberStyle.CIRCLED_SANS_SERIF_NUMBER,
            (
                (1, "➀"),
                (2, "➁"),
                (3, "➂"),
                (4, "➃"),
                (5, "➄"),
                (6, "➅"),
                (7, "➆"),
                (8, "➇"),
                (9, "➈"),
                (10, "➉"),
            ),
        ),
        (
            NumberClass.BLACK_CIRCLED_NUMBER,
            NumberStyle.BLACK_CIRCLED_SANS_SERIF_NUMBER,
            (
                (1, "➊"),
                (2, "➋"),
                (3, "➌"),
                (4, "➍"),
                (5, "➎"),
                (6, "➏"),
                (7, "➐"),
                (8, "➑"),
                (9, "➒"),
                (10, "➓"),
            ),
        ),
        (
            NumberClass.BLACK_CIRCLED_NUMBER,
            NumberStyle.BLACK_CIRCLED_NUMBER,
            (
                (0, "⓿"),
                (1, "❶"),
                (2, "❷"),
                (3, "❸"),
                (4, "❹"),
                (5, "❺"),
                (6, "❻"),
                (7, "❼"),
                (8, "❽"),
                (9, "❾"),
                (10, "❿"),
                (11, "⓫"),
                (12, "⓬"),
                (13, "⓭"),
                (14, "⓮"),
                (15, "⓯"),
                (16, "⓰"),
                (17, "⓱"),
                (18, "⓲"),
                (19, "⓳"),
                (20, "⓴"),
            ),
        ),
        (
            NumberClass.PARENTHESIZED_CHINESE,
            NumberStyle.PARENTHESIZED_CHINESE,
            (
                (1, "㈠"),
                (2, "㈡"),
                (3, "㈢"),
                (4, "㈣"),
                (5, "㈤"),
                (6, "㈥"),
                (7, "㈦"),
                (8, "㈧"),
                (9, "㈨"),
                (10, "㈩"),
            ),
        ),
        (
            NumberClass.CIRCLED_CHINESE,
            NumberStyle.CIRCLED_CHINESE,
            (
                (1, "㊀"),
                (2, "㊁"),
                (3, "㊂"),
                (4, "㊃"),
                (5, "㊄"),
                (6, "㊅"),
                (7, "㊆"),
                (8, "㊇"),
                (9, "㊈"),
                (10, "㊉"),
            ),
        ),
        (
            NumberClass.UNBOUNDED_NUMBER,
            NumberStyle.FULL_WIDTH_NUMBER,
            (
                (0, "０"),
                (1, "１"),
                (2, "２"),
                (3, "３"),
                (4, "４"),
                (5, "５"),
                (6, "６"),
                (7, "７"),
                (8, "８"),
                (9, "９"),
            ),
        ),
        (
            NumberClass.UNBOUNDED_NUMBER,
            NumberStyle.MATHEMATICAL_BOLD_NUMBER,
            (
                (0, "𝟬"),
                (1, "𝟭"),
                (2, "𝟮"),
                (3, "𝟯"),
                (4, "𝟰"),
                (5, "𝟱"),
                (6, "𝟲"),
                (7, "𝟳"),
                (8, "𝟴"),
                (9, "𝟵"),
            ),
        ),
        (
            NumberClass.UNBOUNDED_NUMBER,
            NumberStyle.ARTISTIC_BOLD_NUMBER,
            (
                (0, "𝟎"),
                (1, "𝟏"),
                (2, "𝟐"),
                (3, "𝟑"),
                (4, "𝟒"),
                (5, "𝟓"),
                (6, "𝟔"),
                (7, "𝟕"),
                (8, "𝟖"),
                (9, "𝟗"),
            ),
        ),
        (
            NumberClass.UNBOUNDED_NUMBER,
            NumberStyle.OUTLINED_BOLD_NUMBER,
            (
                (0, "𝟘"),
                (1, "𝟙"),
                (2, "𝟚"),
                (3, "𝟛"),
                (4, "𝟜"),
                (5, "𝟝"),
                (6, "𝟞"),
                (7, "𝟟"),
                (8, "𝟠"),
                (9, "𝟡"),
            ),
        ),
        (
            NumberClass.UNBOUNDED_NUMBER,
            NumberStyle.SUBSCRIPT_NUMBER,
            (
                (0, "🄁"),
                (1, "🄂"),
                (2, "🄃"),
                (3, "🄄"),
                (4, "🄅"),
                (5, "🄆"),
                (6, "🄇"),
                (7, "🄈"),
                (8, "🄉"),
                (9, "🄊"),
            ),
        ),
    )
)



from .handler import DefaultPDFDocument, DefaultPDFHandler, PDFDocument, PDFHandler
from .ocr import OCR, OCREvent, OCREventKind
from .page_ref import pdf_pages_count
from .ref import *
from .types import (
    DeepSeekOCRSize,
    Page,
    PageLayout,
    PDFDocumentMetadata,
    decode,
    encode,
)



from pathlib import Path
from typing import Generator

from ..common import XMLReader, save_xml
from ..pdf import TITLE_TAGS, Page, decode
from ..toc import Toc, TocInfo, iter_toc
from .analyse_level import analyse_chapter_internal_levels
from .chapter import (
    AssetLayout,
    BlockLayout,
    Chapter,
    ParagraphLayout,
    Reference,
    encode,
)
from .content import expand_text_in_content, join_texts_in_content
from .jointer import Jointer
from .mark import Mark, search_marks
from .reference import References


def generate_chapter_files(pages_path: Path, chapters_path: Path, toc: TocInfo):
    chapters_path.mkdir(parents=True, exist_ok=True)
    for chapter_file in chapters_path.glob("chapter_*.xml"):
        chapter_file.unlink()

    for chapter in _generate_chapters(
        pages_path=pages_path,
        toc=toc,
    ):
        tail: str
        if chapter.id is None:
            tail = "head"
        else:
            tail = f"{chapter.id}"

        chapter = analyse_chapter_internal_levels(chapter)
        chapter_file = chapters_path / f"chapter_{tail}.xml"
        chapter_element = encode(chapter)
        save_xml(chapter_element, chapter_file)


def _generate_chapters(
    pages_path: Path, toc: TocInfo
) -> Generator[Chapter, None, None]:
    chapter: Chapter | None = None
    ref2toc: dict[tuple[int, int], Toc] = {}

    for item in iter_toc(toc.content):
        ref2toc[(item.page_index, item.order)] = item

    for layout in _extract_body_layouts(pages_path, toc):
        matched_toc = False
        if (
            isinstance(layout, ParagraphLayout)
            and layout.blocks
            and layout.ref in TITLE_TAGS
        ):
            item: Toc | None = None
            for block in layout.blocks:
                item = ref2toc.get((block.page_index, block.order), None)
                if item:
                    break
            if item:
                if chapter:
                    yield chapter
                chapter = Chapter(
                    id=item.id,
                    level=item.level,
                    layouts=[layout],
                )
                matched_toc = True

        if not matched_toc:
            if chapter is None:
                max_level = max((t.level for t in iter_toc(toc.content)), default=0)
                chapter = Chapter(
                    id=None,
                    level=max_level,  # 防止章节标题盖过其他
                    layouts=[],
                )
            chapter.layouts.append(layout)

    if chapter:
        yield chapter


def _extract_body_layouts(pages_path: Path, toc: TocInfo):
    pages: XMLReader[Page] = XMLReader(
        prefix="page",
        dir_path=pages_path,
        decode=decode,
    )
    toc_page_indexes = set(toc.page_indexes)
    body_jointer = Jointer(
        (
            (p.index, p.body_layouts)
            for p in pages.read()
            if p.index not in toc_page_indexes
        )
    )
    footnotes_jointer = Jointer(
        (
            (p.index, p.footnotes_layouts)
            for p in pages.read()
            if p.index not in toc_page_indexes
        )
    )
    references_generator = _extract_page_references(footnotes_jointer)
    current_references: References | None = next(references_generator, None)

    def get_references(page_index: int) -> References | None:
        nonlocal current_references
        while (
            current_references is not None
            and current_references.page_index < page_index
        ):
            current_references = next(references_generator, None)
        if (
            current_references is not None
            and current_references.page_index == page_index
        ):
            return current_references
        return None

    for layout in body_jointer.execute():
        if isinstance(layout, ParagraphLayout):
            for block in layout.blocks:
                references = get_references(block.page_index)
                if references:
                    _replace_mark_with_reference(references, block)
                join_texts_in_content(block.content)

        yield layout


def _extract_page_references(jointer: Jointer) -> Generator[References, None, None]:
    last_page_index: int = -1
    layout_buffer: list[AssetLayout | ParagraphLayout] = []

    for layout in jointer.execute():
        page_index = _page_index_from_layout(layout)
        if page_index != last_page_index:
            if layout_buffer:
                yield References(
                    page_index=last_page_index,
                    layouts=layout_buffer,
                )
            last_page_index = page_index
            layout_buffer = []
        layout_buffer.append(layout)

    if layout_buffer:
        yield References(
            page_index=last_page_index,
            layouts=layout_buffer,
        )


def _page_index_from_layout(layout: AssetLayout | ParagraphLayout) -> int:
    if isinstance(layout, ParagraphLayout):
        if not layout.blocks:
            raise ValueError("ParagraphLayout has no blocks to get page index")
        return layout.blocks[0].page_index
    elif isinstance(layout, AssetLayout):
        return layout.page_index
    else:
        raise TypeError(f"Unknown layout type: {type(layout).__name__}")


def _replace_mark_with_reference(references: References, block: BlockLayout):
    def expand(text: str):
        for item in search_marks(text):
            reference: Reference | None = None
            if isinstance(item, Mark):
                reference = references.get(item)
            if reference:
                yield reference
            else:
                yield str(item)

    expand_text_in_content(
        content=block.content,
        expand=expand,
    )



from os import PathLike
from pathlib import Path
from typing import Callable, Literal

from epub_generator import BookMeta, LaTeXRender, TableRender

from .common import EnsureFolder, remove_surrogates
from .epub import render_epub_file
from .error import (
    IgnoreOCRErrorsChecker,
    IgnorePDFErrorsChecker,
    PDFError,
    is_inline_error,
    to_interrupted_error,
)
from .markdown.render import render_markdown_file
from .metering import AbortedCheck, OCRTokensMetering
from .pdf import OCR, DeepSeekOCRSize, OCREvent, PDFHandler
from .sequence import generate_chapter_files
from .to_path import to_path
from .toc import analyse_toc


class Transform:
    def __init__(
        self,
        models_cache_path: PathLike | str | None = None,
        pdf_handler: PDFHandler | None = None,
        local_only: bool = False,
    ) -> None:
        self._ocr: OCR = OCR(
            model_path=models_cache_path,
            pdf_handler=pdf_handler,
            local_only=local_only,
        )

    def predownload(self, revision: str | None = None) -> None:
        self._ocr.predownload(revision)

    def load_models(self) -> None:
        self._ocr.load_models()

    def transform_markdown(
        self,
        pdf_path: PathLike | str,
        markdown_path: PathLike | str,
        markdown_assets_path: PathLike | str | None = None,
        analysing_path: PathLike | str | None = None,
        ocr_size: DeepSeekOCRSize = "gundam",
        dpi: int | None = None,
        max_page_image_file_size: int | None = None,
        includes_cover: bool = False,
        includes_footnotes: bool = False,
        generate_plot: bool = False,
        toc_assumed: bool = False,
        ignore_pdf_errors: IgnorePDFErrorsChecker = False,
        ignore_ocr_errors: IgnoreOCRErrorsChecker = False,
        aborted: AbortedCheck = lambda: False,
        max_ocr_tokens: int | None = None,
        max_ocr_output_tokens: int | None = None,
        on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
    ) -> OCRTokensMetering:  # pyright: ignore[reportReturnType]
        if markdown_assets_path is None:
            markdown_assets_path = Path(".") / "assets"
        else:
            markdown_assets_path = Path(markdown_assets_path)
        try:
            with EnsureFolder(
                path=to_path(analysing_path) if analysing_path is not None else None,
            ) as analysing_path:
                asserts_path, chapters_path, _, cover_path, metering = (
                    self._extract_from_pdf(
                        pdf_path=Path(pdf_path),
                        analysing_path=analysing_path,
                        ocr_size=ocr_size,
                        dpi=dpi,
                        max_page_image_file_size=max_page_image_file_size,
                        includes_cover=includes_cover,
                        includes_footnotes=includes_footnotes,
                        ignore_pdf_errors=ignore_pdf_errors,
                        ignore_ocr_errors=ignore_ocr_errors,
                        generate_plot=generate_plot,
                        toc_assumed=toc_assumed,
                        aborted=aborted,
                        max_tokens=max_ocr_tokens,
                        max_output_tokens=max_ocr_output_tokens,
                        on_ocr_event=on_ocr_event,
                    )
                )
                render_markdown_file(
                    chapters_path=chapters_path,
                    assets_path=asserts_path,
                    output_path=Path(markdown_path),
                    output_assets_path=markdown_assets_path,
                    cover_path=cover_path,
                    aborted=aborted,
                )
                return metering

        except Exception as raw_error:
            error = to_interrupted_error(raw_error)
            if error:
                raise error from raw_error
            elif is_inline_error(raw_error):
                raise
            else:
                raise RuntimeError(
                    f"transform {pdf_path} to markdown failed"
                ) from raw_error

    def transform_epub(
        self,
        pdf_path: PathLike | str,
        epub_path: PathLike | str,
        analysing_path: PathLike | str | None = None,
        ocr_size: DeepSeekOCRSize = "gundam",
        dpi: int | None = None,
        max_page_image_file_size: int | None = None,
        includes_cover: bool = True,
        includes_footnotes: bool = False,
        ignore_pdf_errors: IgnorePDFErrorsChecker = False,
        ignore_ocr_errors: IgnoreOCRErrorsChecker = False,
        generate_plot: bool = False,
        toc_assumed: bool = True,
        book_meta: BookMeta | None = None,
        lan: Literal["zh", "en"] = "zh",
        table_render: TableRender = TableRender.HTML,
        latex_render: LaTeXRender = LaTeXRender.MATHML,
        inline_latex: bool = True,
        aborted: AbortedCheck = lambda: False,
        max_ocr_tokens: int | None = None,
        max_ocr_output_tokens: int | None = None,
        on_ocr_event: Callable[[OCREvent], None] = lambda _: None,
    ) -> OCRTokensMetering:  # pyright: ignore[reportReturnType]
        try:
            with EnsureFolder(
                path=to_path(analysing_path) if analysing_path is not None else None,
            ) as analysing_path:
                pdf_path = Path(pdf_path)
                asserts_path, chapters_path, toc_path, cover_path, metering = (
                    self._extract_from_pdf(
                        pdf_path=pdf_path,
                        analysing_path=analysing_path,
                        ocr_size=ocr_size,
                        dpi=dpi,
                        max_page_image_file_size=max_page_image_file_size,
                        includes_cover=includes_cover,
                        includes_footnotes=includes_footnotes,
                        ignore_pdf_errors=ignore_pdf_errors,
                        ignore_ocr_errors=ignore_ocr_errors,
                        generate_plot=generate_plot,
                        toc_assumed=toc_assumed,
                        aborted=aborted,
                        max_tokens=max_ocr_tokens,
                        max_output_tokens=max_ocr_output_tokens,
                        on_ocr_event=on_ocr_event,
                    )
                )
                book_meta = book_meta or self._extract_book_meta(pdf_path)

                render_epub_file(
                    chapters_path=chapters_path,
                    toc_path=toc_path,
                    assets_path=asserts_path,
                    epub_path=Path(epub_path),
                    book_meta=book_meta,
                    lan=lan,
                    cover_path=cover_path,
                    table_render=table_render,
                    latex_render=latex_render,
                    inline_latex=inline_latex,
                    aborted=aborted,
                )
                return metering

        except Exception as raw_error:
            error = to_interrupted_error(raw_error)
            if error:
                raise error from raw_error
            elif is_inline_error(raw_error):
                raise
            else:
                raise RuntimeError(
                    f"transform {pdf_path} to epub failed"
                ) from raw_error

    def _extract_from_pdf(
        self,
        pdf_path: Path,
        analysing_path: Path,
        ocr_size: DeepSeekOCRSize,
        dpi: int | None,
        max_page_image_file_size: int | None,
        includes_cover: bool,
        includes_footnotes: bool,
        ignore_pdf_errors: IgnorePDFErrorsChecker,
        ignore_ocr_errors: IgnoreOCRErrorsChecker,
        generate_plot: bool,
        toc_assumed: bool,
        aborted: AbortedCheck,
        max_tokens: int | None,
        max_output_tokens: int | None,
        on_ocr_event: Callable[[OCREvent], None],
    ):
        asserts_path = analysing_path / "assets"
        pages_path = analysing_path / "ocr"
        chapters_path = analysing_path / "chapters"
        toc_path = analysing_path / "toc.xml"

        cover_path: Path | None = None
        plot_path: Path | None = None
        if includes_cover:
            cover_path = analysing_path / "cover.png"
        if generate_plot:
            plot_path = analysing_path / "plots"

        metering = OCRTokensMetering(
            input_tokens=0,
            output_tokens=0,
        )
        for event in self._ocr.recognize(
            pdf_path=pdf_path,
            asset_path=asserts_path,
            ocr_path=pages_path,
            ocr_size=ocr_size,
            dpi=dpi,
            max_page_image_file_size=max_page_image_file_size,
            includes_footnotes=includes_footnotes,
            ignore_pdf_errors=ignore_pdf_errors,
            ignore_ocr_errors=ignore_ocr_errors,
            plot_path=plot_path,
            cover_path=cover_path,
            aborted=aborted,
            max_tokens=max_tokens,
            max_output_tokens=max_output_tokens,
        ):
            on_ocr_event(event)
            metering.input_tokens += event.input_tokens
            metering.output_tokens += event.output_tokens

        toc = analyse_toc(
            pages_path=pages_path,
            toc_path=toc_path,
            toc_assumed=toc_assumed,
        )
        generate_chapter_files(
            pages_path=pages_path,
            chapters_path=chapters_path,
            toc=toc,
        )
        if cover_path and not cover_path.exists():
            cover_path = None

        return asserts_path, chapters_path, toc_path, cover_path, metering

    def _extract_book_meta(self, pdf_path: Path) -> BookMeta | None:
        try:
            pdf_metadata = self._ocr.metadata(pdf_path)
            return BookMeta(
                title=self._normalize_text_in_meta(pdf_metadata.title) or pdf_path.stem,
                description=self._normalize_text_in_meta(pdf_metadata.description),
                publisher=self._normalize_text_in_meta(pdf_metadata.publisher),
                isbn=self._normalize_text_in_meta(pdf_metadata.isbn),
                authors=[remove_surrogates(s) for s in pdf_metadata.authors],
                editors=[remove_surrogates(s) for s in pdf_metadata.editors],
                translators=[remove_surrogates(s) for s in pdf_metadata.translators],
                modified=pdf_metadata.modified,
            )
        except PDFError:
            print("Warning: Failed to extract PDF metadata.")
            return None

    def _normalize_text_in_meta(self, text: str | None) -> str | None:
        if text is None:
            return None
        return remove_surrogates(text)



from ..common import median, split_by_cv
from ..config import MAX_TITLE_CV
from ..pdf import TITLE_TAGS
from .chapter import Chapter, ParagraphLayout

# markdown 最大支持 6 级标题，减去作为标题的 1 级
_MAX_TITLE_GROUP = 5


def analyse_chapter_internal_levels(chapter: Chapter) -> Chapter:
    for level, layouts in enumerate(
        start=1,  # 0 作为 chapter 总标题保留，故从 1 开始
        iterable=reversed(
            split_by_cv(  # 标题从大到小排列等级，故反转
                payload_items=_collect_heights(chapter),
                max_cv=MAX_TITLE_CV,
                max_groups=_MAX_TITLE_GROUP,
            )
        ),
    ):
        for layout in layouts:
            layout.level = level

    return chapter


def _collect_heights(chapter: Chapter):
    layout_items: list[tuple[float, ParagraphLayout]] = []
    for i, layout in enumerate(chapter.layouts):
        if not isinstance(layout, ParagraphLayout) or layout.ref not in TITLE_TAGS:
            continue
        if i == 0:
            layout.level = 0
        elif layout.blocks:
            height = median(b.det[3] - b.det[1] for b in layout.blocks)
            layout_items.append((height, layout))
    return layout_items



from dataclasses import dataclass
from enum import Enum, auto
from typing import Callable

AbortedCheck = Callable[[], bool]


def check_aborted(aborted_check: AbortedCheck) -> None:
    if aborted_check():
        from doc_page_extractor import AbortError

        raise AbortError()


@dataclass
class OCRTokensMetering:
    input_tokens: int
    output_tokens: int


class InterruptedKind(Enum):
    ABORT = auto()
    TOKEN_LIMIT_EXCEEDED = auto()
